#include <sys/types.h>
#include <sys/stat.h>
#include "byte.h"
#include "error.h"
#include "exit.h"
#include "fmt.h"
#include "now.h"
#include "open.h"
#include "readwrite.h"
#include "sig.h"
#include "str.h"
#include "strerr.h"
#include "substdio.h"

#define FATAL "maildirdeliver: fatal: "
#define PERM 100
#define TEMP 111

extern int alarm();
extern int chdir();
extern int close();
extern int fsync();
extern int gethostname();
extern int getpid();
extern int link();
extern int sleep();
extern int unlink();

static char inbuf[1024];
static char outbuf[1024];

static char fntmptph[80 + FMT_ULONG * 2];
static char fnnewtph[80 + FMT_ULONG * 2];

static void exitnicely (int code)
{
    unlink(fntmptph);
    _exit(code);
}

static void sigalrm (void)
{
    strerr_warn2(FATAL, "timeout on delivery", (struct strerr *) 0);
    exitnicely(TEMP);
}

static void die_read (void)
{
    strerr_warn2(FATAL, "unable to read message: ", &strerr_sys);
    exitnicely(TEMP);
}

int main (int argc, char **argv)
{
    char *dir;
    int code;
    unsigned long pid;
    unsigned long time;
    char host[64];
    char *s;
    int loop;
    struct stat st;
    int fd;
    substdio ss;
    substdio ssout;
    int r;
    char c;

    ++argv;

    dir = *argv++;
    if (!dir) strerr_die1x(PERM, "maildirdeliver: usage: maildirdeliver dir");

    umask(077);
    sig_alarmcatch(sigalrm);

    if (chdir(dir) == -1) {
	code = error_temp(errno) ? TEMP : PERM;
	strerr_die4sys(code, FATAL, "unable to chdir to ", dir, ": ");
    }

    pid = getpid();
    host[0] = '\0';
    gethostname(host, sizeof(host));
    for (loop = 0;; ++loop) {
	time = now();
	s = fntmptph;
	s += fmt_str(s, "tmp/");
	s += fmt_ulong(s, time); *s++ = '.';
	s += fmt_ulong(s, pid); *s++ = '.';
	s += fmt_strn(s, host, sizeof(host)); *s++ = 0;
	if (stat(fntmptph, &st) == -1) if (errno == error_noent) break;
	if (loop == 2)
	    strerr_die2x(TEMP, FATAL, "unable to get temporary file name");
	sleep(2);
    }
    str_copy(fnnewtph, fntmptph);
    byte_copy(fnnewtph, 3, "new");

    alarm(86400);

    fd = open_excl(fntmptph);
    if (fd == -1)
	strerr_die4sys(TEMP, FATAL, "unable to open ", fntmptph, ": ");
    substdio_fdbuf(&ss, read, 0, inbuf, sizeof(inbuf));
    substdio_fdbuf(&ssout, write, fd, outbuf, sizeof(outbuf));

    /* discard initial From line, if any. */
    for (;;) {
	r = substdio_feed(&ss);
	if (r == -1) die_read();
	if (r == 0) break;
	if (r >= 5) break;
    }
    if ((r >= 5) && (byte_equal(substdio_PEEK(&ss), 5, "From "))) {
	for (;;) {
	    r = substdio_get(&ss, &c, 1);
	    if (r == -1) die_read();
	    if (r == 0) break;
	    if (c == '\n') break;
	}
    }

    switch (substdio_copy(&ssout, &ss)) {
    case -2:
	die_read();
    case -3:
	strerr_warn2(FATAL, "unable to write message: ", &strerr_sys);
	exitnicely(TEMP);
    }

    r = substdio_flush(&ssout);
    if (r != -1) r = fsync(fd);
    if (r != -1) r = close(fd);
    if (r == -1) {
	strerr_warn2(FATAL, "unable to write message: ", &strerr_sys);
	exitnicely(TEMP);
    }

    if (link(fntmptph, fnnewtph) == -1) {
	strerr_warn4(FATAL, "unable to link ", fnnewtph, ": ", &strerr_sys);
	exitnicely(TEMP);
    }

    exitnicely(0);
    return 0; /* shut up gcc -Wall */
}
